IMPLEMENTATION MODULE BTREES;

FROM LOWFILES IMPORT FilePos, RewriteLowFile, ResetLowFile, CloseLowFile,
                     ReadNBytes, WriteNBytes;
FROM Storage IMPORT ALLOCATE, DEALLOCATE, Available;
FROM SYSTEM IMPORT TSIZE, ADR, ADDRESS, BYTE;

PROCEDURE Move(Von, Nach : ADDRESS; Size : CARDINAL);
  (* Kann durch eventuell vorhandene Bibliotheksroutine ersetzt werden *)
  VAR v,n : POINTER TO ARRAY[0..TSIZE(Item)] OF BYTE;
  BEGIN
    v:=Von; n:=Nach;
    WHILE Size>0 DO
      DEC(Size);
      n^[Size]:=v^[Size];
    END
  END Move;

PROCEDURE AssignKey(VAR B : BTree; VAR s,d : ARRAY OF BYTE);
  (* Tritt an Stelle der normalen Zuweisung. Aus diesem Grund mu die
     bergebene KeySize immer kleiner oder gleich TSIZE(Key) sein! *)
  VAR i : CARDINAL;
  BEGIN
    Move(ADR(s),ADR(d),B.KeySize)
  END AssignKey;

PROCEDURE ClearPath(VAR B : BTree; VAR IO : INTEGER);
  (* Reinigt den Suchpfad *)
  VAR i : CARDINAL;
  BEGIN
    IO:=0;
    WITH B DO
      FOR i:=1 TO Height DO
        WITH Path[i] DO
          IF (MPtr<>NIL) AND NOT UnChanged THEN
            WriteNBytes(f,VAL(FilePos,Way)*VAL(FilePos,PageSize),
                          PageSize,MPtr^,IO)
          END; (* IF *)
          IF IO<>0 THEN RETURN END;
          IF MPtr<>NIL THEN
            DEALLOCATE(MPtr,PageSize); MPtr:=NIL; Way:=0
          END (* IF *)
        END (* WITH *)
      END (* FOR *)
    END (* WITH *)
  END ClearPath;

PROCEDURE Allocate(VAR p : ADDRESS; Size : CARDINAL; VAR Okay : BOOLEAN);
  (* berwachte Speicherreservierung *)
  BEGIN
    Okay:=Available(Size+10000);
    IF Okay THEN ALLOCATE(p,Size) ELSE p:=NIL END
  END Allocate;

PROCEDURE GetItem(VAR b : BTree; VAR p : Page; x : CARDINAL;
                  VAR v : ARRAY OF BYTE);
  (* Holt den x-ten Item aus der Seite p in die Variable v *)
  BEGIN
    Move(ADR(p.e[(x-1)*b.ItemSize]),ADR(v),b.ItemSize);
  END GetItem;

PROCEDURE PutItem(VAR b : BTree; VAR p : Page; x : CARDINAL;
                  VAR v : ARRAY OF BYTE);
  (* Trgt den x-ten Item v in die Seite p ein *)
  BEGIN
    Move(ADR(v),ADR(p.e[(x-1)*b.ItemSize]),b.ItemSize);
  END PutItem;

PROCEDURE GetPage(VAR b : BTree; pn,h : CARDINAL; VAR IO : INTEGER);
  (* Holt eine Seite *)
  VAR Okay : BOOLEAN;
  BEGIN
    IO:=0;
    WITH b DO
      WITH Path[h] DO
        IF h=0
        THEN ReadNBytes(f,VAL(FilePos,pn)*VAL(FilePos,PageSize),
                          PageSize,MPtr^,IO)
        ELSIF Way<>pn (* ansonsten ist die Seite schon da *)
        THEN IF MPtr=NIL
             THEN
               Allocate(MPtr,PageSize,Okay);
               IF NOT Okay THEN IO:=-1 END
             ELSIF NOT UnChanged THEN
               WriteNBytes(f,VAL(FilePos,Way)*VAL(FilePos,PageSize),
                             PageSize,MPtr^,IO)
             END; (* Jetzt steht eine freie Seite bereit *)
             IF IO=0 THEN
               ReadNBytes(f,VAL(FilePos,pn)*VAL(FilePos,PageSize),
                            PageSize,MPtr^,IO)
             END;
             Way:=pn; UnChanged:=TRUE
        END (* IF Way<>pn *)
      END
    END (* WITH b *)
  END GetPage;

PROCEDURE PutPage(VAR b : BTree; pn,h : CARDINAL; VAR IO : INTEGER);
  (* Schreibt eine Seite zurck *)
  VAR Okay : BOOLEAN;
  BEGIN
    IO:=0;
    WITH b DO
     WITH Path[h] DO
       IF h=0
       THEN WriteNBytes(f,VAL(FilePos,pn)*VAL(FilePos,PageSize),
                          PageSize,MPtr^,IO)
       ELSIF Way<>pn
       THEN IF MPtr=NIL
            THEN
              Allocate(MPtr,PageSize,Okay);
              IF NOT Okay THEN IO:=-1 END
            ELSE
              WriteNBytes(f,VAL(FilePos,Way)*VAL(FilePos,PageSize),
                            PageSize,MPtr^,IO)
            END;
            Way:=pn; UnChanged:=FALSE;
       END
     END
    END (* WITH *)
  END PutPage;

PROCEDURE PageInd(VAR b : BTree; VAR p : Page; VAR k : ARRAY OF BYTE;
                 VAR found : BOOLEAN) : CARDINAL;
  (* Gibt die Stelle an, an der k einzufgen wre. Falls found=TRUE,
     dann ist k wirklich vohanden. *)
  VAR i,l,r : CARDINAL;
  BEGIN
    WITH b DO
      WITH p DO
        l:=1; r:=m; found:=FALSE;
        WHILE l<=r DO (* Binre Suche *)
          i:=(l+r) DIV 2; GetItem(b,p,i,w^);
          IF Relation(ADR(w^.Key),ADR(k)) THEN l:=i+1
          ELSIF Relation(ADR(k),ADR(w^.Key)) THEN r:=i-1
          ELSE found:=TRUE; RETURN i
          END
        END (* WHILE *)
      END
    END; (* WITH *)
    RETURN r;
  END PageInd;

PROCEDURE SearchNode(VAR b : BTree; VAR pn,h,r : CARDINAL;
                     VAR k : ARRAY OF BYTE; VAR found : BOOLEAN);
  (* Sucht den Schlssel k und baut dabei den Suchpfad bis zur Fundstelle
     auf. Falls found=TRUE, dann wurde der Eintrag gefunden. Anderfalls
     zeigt pn auf die Blattseite, in der k an der r-ten Stelle eingefgt
     werden mte. In diesem Fall ist der Pfad vollstndig und h entspricht
     der Baumhhe. *)
  VAR p ,IO : INTEGER;
  BEGIN
    WITH b DO
      r:=0; h:=1; p:=1; found:=FALSE;
      WHILE h<=Height DO
        pn:=p; GetPage(b,p,h,IO); IF IO<>0 THEN RETURN END;
        WITH Path[h] DO
          WITH MPtr^ DO
            r:=PageInd(b,MPtr^,k,found); IF found THEN RETURN END;
            IF r>0 THEN GetItem(b,MPtr^,r,w^) END;
            INC(h); IF r=0 THEN p:=p0 ELSE p:=w^.p END
          END
        END (* WITH *)
      END (* WHILE *)
    END; (* WITH b *)
    DEC(h);
  END SearchNode;

PROCEDURE first(VAR b : BTree; pn,h : CARDINAL; VAR k : ARRAY OF BYTE;
                VAR found : BOOLEAN);
 (* Sucht den ersten Eintrag des durch pn (und h) bestimmten Teilbaums.
    Da sich dieser Eintrag grundstzlich auf einer Blattseite befindet,
    wird von h ab der Suchpfad komplett aufgebaut *)
  VAR ready : BOOLEAN;
      IO : INTEGER;
  BEGIN
    WITH b DO
      found:=FALSE;
      REPEAT
        GetPage(b,pn,h,IO); IF IO<>0 THEN RETURN END;
        WITH Path[h].MPtr^ DO
          ready:=p0=0; IF NOT ready THEN pn:=p0; INC(h) END
        END
      UNTIL ready;
      WITH Path[h] DO
        WITH MPtr^ DO
          IF m>0
          THEN
            GetItem(b,MPtr^,1,w^); AssignKey(b,w^.Key,k); found:=TRUE
          END (* IF *)
        END
      END
    END (* WITH *)
  END first;

PROCEDURE last(VAR b : BTree; pn,h : CARDINAL; VAR k : ARRAY OF BYTE;
               VAR found : BOOLEAN);
 (* Sucht den letzten Eintrag des durch pn (und h) bestimmten Teilbaums.
    Da sich dieser Eintrag grundstzlich auf einer Blattseite befindet,
    wird von h ab der Suchpfad komplett aufgebaut *)
  VAR ready : BOOLEAN;
      IO : INTEGER;
  BEGIN
    WITH b DO
      found:=FALSE;
      REPEAT
        GetPage(b,pn,h,IO); IF IO<>0 THEN RETURN END;
        WITH Path[h] DO
          ready:=MPtr^.p0=0;
          IF NOT ready
          THEN GetItem(b,MPtr^,MPtr^.m,w^); pn:=w^.p; INC(h) END
        END
      UNTIL ready;
      WITH Path[h] DO
        WITH MPtr^ DO
          IF m>0
          THEN
            GetItem(b,MPtr^,m,w^); AssignKey(b,w^.Key,k); found:=TRUE
          END
        END
      END
    END (* WITH *)
  END last;

(**************************************************************************)

PROCEDURE NewTree(VAR B : BTree; FileName : Str64; ps,ks : CARDINAL;
                  Rel : Vergleich; VAR IO : INTEGER);
  VAR i : CARDINAL;
      Okay : BOOLEAN;
  BEGIN
    WITH B DO
      RewriteLowFile(f,FileName,IO); IF IO<>0 THEN RETURN END;
      Height:=0; KeySize:=ks; ItemSize:=ks+TSIZE(CARDINAL);
      PageSize:=ps; n:=((PageSize-2*TSIZE(CARDINAL)) DIV ItemSize) DIV 2;
      n2:=2*n; Pages:=0; Relation:=Rel; IF n=0 THEN IO:=-1; RETURN END;
      Allocate(u,ItemSize,Okay); IF NOT Okay THEN IO:=-2; RETURN END;
      Allocate(v,ItemSize,Okay); IF NOT Okay THEN IO:=-2; RETURN END;
      Allocate(w,ItemSize,Okay); IF NOT Okay THEN IO:=-2; RETURN END;
      WITH Path[0] DO
        Way:=0; Allocate(MPtr,PageSize,Okay);
        IF NOT Okay THEN IO:=-2; RETURN END;
        WriteNBytes(f,VAL(FilePos,0),PageSize,MPtr^,IO);
        IF IO<>0 THEN RETURN END;
        MPtr^.p0:=0; MPtr^.m:=0;
        WriteNBytes(f,VAL(FilePos,PageSize),PageSize,MPtr^,IO);
        IF IO<>0 THEN RETURN END;
      END; (* WITH *)
      FOR i:=1 TO MaxHeight DO
        WITH Path[i] DO Way:=0; MPtr:=NIL END
      END; (* FOR *)
      Relation:=Rel;
      WriteNBytes(f,VAL(FilePos,0),7*TSIZE(CARDINAL),B,IO)
    END (* WITH *)
  END NewTree;

PROCEDURE OpenTree(VAR B : BTree; FileName : Str64; Rel : Vergleich;
                   VAR IO : INTEGER);
VAR i : CARDINAL;
      Okay : BOOLEAN;
  BEGIN
    WITH B DO
      ResetLowFile(f,FileName,IO); IF IO<>0 THEN RETURN END;
      ReadNBytes(f,VAL(FilePos,0),7*TSIZE(CARDINAL),B,IO);
      IF IO<>0 THEN RETURN END;
      Allocate(u,ItemSize,Okay); IF NOT Okay THEN IO:=-2; RETURN END;
      Allocate(v,ItemSize,Okay); IF NOT Okay THEN IO:=-2; RETURN END;
      Allocate(w,ItemSize,Okay); IF NOT Okay THEN IO:=-2; RETURN END;
      WITH Path[0] DO Way:=0; Allocate(MPtr,PageSize,Okay) END;
      IF NOT Okay THEN IO:=-2; RETURN END;
      FOR i:=1 TO MaxHeight DO
        WITH Path[i] DO Way:=0; MPtr:=NIL END;
      END; (* FOR *)
      Relation:=Rel
    END (* WITH *)
  END OpenTree;

PROCEDURE CloseTree(VAR B : BTree; VAR IO : INTEGER);
  BEGIN
    WITH B DO
      WriteNBytes(f,VAL(FilePos,0),7*TSIZE(CARDINAL),B,IO);
      IF IO<>0 THEN RETURN END;
      ClearPath(B,IO); IF IO>0 THEN RETURN END;
      CloseLowFile(f,IO); IF IO<>0 THEN RETURN END;
      WITH Path[0] DO DEALLOCATE(MPtr,PageSize) END;
      DEALLOCATE(u,ItemSize); DEALLOCATE(v,ItemSize); DEALLOCATE(w,ItemSize);
      END (* WITH *)
  END CloseTree;

PROCEDURE InsertKey(VAR b : BTree; VAR k : ARRAY OF BYTE; VAR Okay : BOOLEAN);
  VAR q,p : PagePtr;
      r,i,pn,h  : CARDINAL;
      OverFlow, exists : BOOLEAN;
      IO : INTEGER;
  BEGIN
    Okay:=TRUE;
    WITH b DO
      AssignKey(b,k,u^.Key); u^.p:=0;
      IF Height=0 THEN (* Neuer Baum *)
        Height:=1; Pages:=1;
        WITH Path[0] DO
          WITH MPtr^ DO m:=1; p0:=0; PutItem(b,MPtr^,1,u^) END
        END; (* WITH *)
        PutPage(b,1,0,IO); IF IO<>0 THEN Okay:=FALSE; RETURN END;
        RETURN
      END; (* IF *)
      SearchNode(b,pn,h,r,k,exists);
      IF exists THEN Okay:=FALSE; RETURN END;
      p:=Path[h].MPtr; q:=Path[0].MPtr;
      REPEAT
        OverFlow:=FALSE; Path[h].UnChanged:=FALSE; (* Zum Zurckschreiben *)
        IF p^.m<n2 (* Eintrag pat noch auf die Seite *)
        THEN
          INC(p^.m);
          FOR i:=p^.m TO r+2 BY -1 DO
            GetItem(b,p^,i-1,w^); PutItem(b,p^,i,w^);
          END; (* FOR *)
          PutItem(b,p^,r+1,u^);
        ELSE (* aufteilen auf p^ und q^ *)
          IF r<=n
          THEN (* im linken Feld *)
            IF r=n
            THEN v^.p:=u^.p; AssignKey(b,u^.Key,v^.Key)
            ELSE
              GetItem(b,p^,n,v^);
              FOR i:=n TO r+2 BY -1 DO
                GetItem(b,p^,i-1,w^); PutItem(b,p^,i,w^)
              END; (* FOR *)
              PutItem(b,p^,r+1,u^);
            END; (* IF *)
            FOR i:=1 TO n DO
              GetItem(b,p^,i+n,w^); PutItem(b,q^,i,w^)
            END (* FOR *)
          ELSE (* im rechten Feld *)
            r:=r-n; GetItem(b,p^,n+1,v^);
            FOR i:=1 TO r-1 DO
              GetItem(b,p^,i+n+1,w^); PutItem(b,q^,i,w^)
            END; (* FOR *)
            PutItem(b,q^,r,u^);
            FOR i:=r+1 TO n DO
              GetItem(b,p^,i+n,w^); PutItem(b,q^,i,w^)
            END (* FOR *)
          END; (* aufteilen *)
          INC(Pages); p^.m:=n; q^.m:=n; q^.p0:=v^.p; v^.p:=Pages;
          u^.p:=v^.p; AssignKey(b,v^.Key,u^.Key);
          PutPage(b,Pages,0,IO); (* q wird neu geschrieben *)
          IF IO<>0 THEN Okay:=FALSE; RETURN END;
          IF h=1 (* neue Wurzel *)
          THEN
            INC(Height); INC(Pages); PutPage(b,Pages,1,IO);
            IF IO<>0 THEN Okay:=FALSE; RETURN END;
            PutItem(b,q^,1,u^); q^.p0:=Pages; q^.m:=1; PutPage(b,1,0,IO);
            IF IO<>0 THEN Okay:=FALSE; RETURN END;
           ELSE (* mittleren Knoten nach oben geben *)
             OverFlow:=TRUE; DEC(h);
             p:=Path[h].MPtr; r:=PageInd(b,p^,u^.Key,exists)
           END (* IF h=1 *)
        END (* IF *)
      UNTIL NOT OverFlow
    END (* WITH b *)
  END InsertKey;

PROCEDURE DeleteKey(VAR b : BTree; VAR k : ARRAY OF BYTE; VAR Okay : BOOLEAN);
  VAR q,p,o : PagePtr;
      r,i,pn,h,t  : CARDINAL;
      UnderFlow, exists, swap : BOOLEAN;
      IO : INTEGER;
  BEGIN
    Okay:=TRUE;
    WITH b DO
      SearchNode(b,pn,h,r,k,exists);
      IF NOT exists THEN Okay:=FALSE; RETURN END;
      Okay:=TRUE; Path[h].UnChanged:=FALSE; p:=Path[h].MPtr;
      IF h<Height (* keine BlattSeite *)
      THEN (* vertausche mit kleinstem rechten Nachfolger *)
        GetItem(b,p^,r,u^); first(b,u^.p,h+1,u^.Key,exists);
        IF NOT exists THEN Okay:=FALSE; RETURN END;
        PutItem(b,p^,r,u^); h:=Height; p:=Path[h].MPtr; r:=1;
      ELSE GetItem(b,p^,r,u^); (* in u^.Key ist der zu lschende Key *)
      END;
      DEC(p^.m); Path[h].UnChanged:=FALSE;
      FOR i:=r TO p^.m DO GetItem(b,p^,i+1,w^); PutItem(b,p^,i,w^) END;
      UnderFlow:=(p^.m<n) AND (h>1);
      WHILE UnderFlow DO
        o:=Path[h-1].MPtr; r:=PageInd(b,o^,u^.Key,exists); swap:=FALSE;
        IF r=0
        THEN swap:=TRUE; GetItem(b,o^,1,w^); pn:=w^.p; r:=1
        ELSIF r=1 THEN pn:=o^.p0
        ELSE GetItem(b,o^,r-1,w^); pn:=w^.p
        END; (* IF r=0 *)
        GetPage(b,pn,0,IO); IF IO<>0 THEN Okay:=FALSE; RETURN END;
        q:=Path[0].MPtr;
        IF p^.m+q^.m>=n2
        THEN (* Ausgleichen von p und q *)
          DEC(q^.m); INC(p^.m);
          IF swap
          THEN
            t:=q^.p0; GetItem(b,q^,1,u^); q^.p0:=u^.p;
            GetItem(b,o^,r,v^); u^.p:=v^.p; v^.p:=t;
            FOR i:=1 TO q^.m DO
              GetItem(b,q^,i+1,w^); PutItem(b,q^,i,w^)
            END;
            PutItem(b,o^,r,u^); PutItem(b,p^,p^.m,v^)
          ELSE
            GetItem(b,q^,q^.m+1,u^); GetItem(b,o^,r,v^);
            t:=p^.p0; p^.p0:=u^.p; u^.p:=v^.p; v^.p:=t;
            FOR i:=p^.m TO 2 BY -1 DO
              GetItem(b,p^,i-1,w^); PutItem(b,p^,i,w^);
            END; (* FOR *)
            PutItem(b,o^,r,u^); PutItem(b,p^,1,v^)
          END; (* IF swap *)
          UnderFlow:=FALSE; PutPage(b,pn,0,IO); (* q zurckschreiben *)
          IF IO<>0 THEN Okay:=FALSE; RETURN END;
          Path[h].UnChanged:=FALSE; Path[h-1].UnChanged:=FALSE
        ELSE (* Mischen     *)
          IF swap
          THEN
            GetItem(b,o^,r,u^); u^.p:=q^.p0; PutItem(b,p^,p^.m+1,u^);
            FOR i:=1 TO q^.m DO
              GetItem(b,q^,i,w^); PutItem(b,p^,p^.m+1+i,w^)
            END; (* FOR *)
            p^.m:=1+p^.m+q^.m;
          ELSE
            GetItem(b,o^,r,v^); v^.p:=p^.p0; PutItem(b,q^,q^.m+1,v^);
            FOR i:=1 TO p^.m DO
              GetItem(b,p^,i,w^); PutItem(b,q^,q^.m+1+i,w^)
            END; (* FOR *)
            q^.m:=1+q^.m+p^.m; PutPage(b,pn,0,IO);
          END; (* IF swap *)
          DEC(o^.m);
          FOR i:=r TO o^.m DO
            GetItem(b,o^,i+1,w^); PutItem(b,o^,i,w^)
          END; (* FOR *)
          Path[h].UnChanged:=FALSE; Path[h-1].UnChanged:=FALSE;
          IF o^.m<n
          THEN h:=h-1; p:=o; UnderFlow:=h>1
          ELSE UnderFlow:=FALSE
          END;
          IF Path[1].MPtr^.m=0 THEN (* Baum schrumpft *)
            t:=Path[1].MPtr^.p0; ClearPath(b,IO);
            IF IO<>0 THEN RETURN END;
            GetPage(b,t,1,IO); IF IO<>0 THEN Okay:=FALSE; RETURN END;
            PutPage(b,1,1,IO); IF IO<>0 THEN Okay:=FALSE; RETURN END;
            DEC(Height);
          END
        END (* IF *)
      END (* WHILE *)
    END (* WITH *)
  END DeleteKey;

PROCEDURE FirstKey(VAR B : BTree; VAR Key : ARRAY OF BYTE; VAR Okay : BOOLEAN);
  (* Sucht den ersten Eintrag im Baum *)
  BEGIN
    first(B,1,1,Key,Okay)
  END FirstKey;

PROCEDURE LastKey(VAR B : BTree; VAR Key : ARRAY OF BYTE; VAR Okay : BOOLEAN);
  (* Sucht den letzten Eintrag im Baum *)
  BEGIN
    last(B,1,1,Key,Okay)
  END LastKey;

PROCEDURE NextKey(VAR b : BTree; VAR Key : ARRAY OF BYTE; VAR Okay : BOOLEAN);
   (* Es wird der vorhergehende Schlssel gefunden. Ist Okay=FALSE,
      dann wird Key nicht verndert. *)
   VAR p : PagePtr;
       pn, h, r : CARDINAL;
   BEGIN
     WITH b DO
       SearchNode(b,pn,h,r,Key,Okay); IF NOT Okay THEN RETURN END;
       p:=Path[h].MPtr;
       GetItem(b,p^,r,w^);
       IF w^.p=0 (* BlattSeite *)
       THEN IF r<p^.m
            THEN INC(r); GetItem(b,p^,r,w^)
            ELSE REPEAT
                   IF h=1
                   THEN r:=0
                   ELSE
                     DEC(h); p:=Path[h].MPtr;
                     r:=1+PageInd(b,p^,Key,Okay);
                     IF r<=p^.m THEN GetItem(b,p^,r,w^) ELSE r:=0 END
                   END
                 UNTIL (h=1) OR (r>0)
            END
       ELSE
         pn:=w^.p; INC(h); first(b,pn,h,Key,Okay); AssignKey(b,Key,w^.Key)
       END; (* END *)
       IF r<>0
       THEN AssignKey(b,w^.Key,Key); Okay:=TRUE
       ELSE Okay:=FALSE
       END
     END (* WITH *)
   END NextKey;

PROCEDURE PrevKey(VAR b : BTree; VAR Key : ARRAY OF BYTE; VAR Okay : BOOLEAN);
   (* Es wird der vorhergehende Schlssel gefunden. Ist Okay=FALSE,
      dann wird Key nicht verndert. *)
   VAR p : PagePtr;
       pn, h, r : CARDINAL;
   BEGIN
     WITH b DO
       SearchNode(b,pn,h,r,Key,Okay); IF NOT Okay THEN RETURN END;
       p:=Path[h].MPtr;
       IF p^.p0=0 (* BlattSeite *)
       THEN IF r>1
            THEN DEC(r); GetItem(b,p^,r,w^)
            ELSE REPEAT
                   IF h=1
                   THEN r:=0
                   ELSE
                     DEC(h); p:=Path[h].MPtr;
                     r:=PageInd(b,p^,Key,Okay);
                     IF r>0 THEN GetItem(b,p^,r,w^) END
                   END
                 UNTIL (h=1) OR (r>0)
            END
       ELSE
         DEC(r);
         IF r=0
         THEN pn:=p^.p0; r:=1
         ELSE GetItem(b,p^,r,w^); pn:=w^.p
         END;
         INC(h); last(b,pn,h,Key,Okay); AssignKey(b,Key,w^.Key);
       END;
       IF r<>0
       THEN AssignKey(b,w^.Key,Key); Okay:=TRUE
       ELSE Okay:=FALSE
       END
     END (* WITH *)
   END PrevKey;

PROCEDURE FindKey(VAR B : BTree; VAR Key : ARRAY OF BYTE; VAR Okay : BOOLEAN);
  (* Falls nach der Suche Okay=TRUE, dann wurde ein passender Eintrag
     gefunden. Ansonsten wird der nchste Eintrag zurckgegeben *)
  VAR pn, h, r : CARDINAL;
  BEGIN
    SearchNode(B,pn,h,r,Key,Okay);
    IF NOT Okay
    THEN
      WITH B DO
        WITH Path[h] DO
          WITH MPtr^ DO
            IF r<m
            THEN INC(r); GetItem(B,MPtr^,r,B.w^); AssignKey(B,w^.Key,Key)
            ELSIF r<>0
            THEN
              GetItem(B,MPtr^,r,B.w^); AssignKey(B,w^.Key,Key);
              NextKey(B,Key,Okay); Okay:=FALSE
            END (* IF *)
          END
        END
      END (* WITH *)
    END (* IF *)
  END FindKey;

PROCEDURE Height(VAR B : BTree):CARDINAL;  (* Aktuelle Hhe eines Baumes *)
  BEGIN
    RETURN B.Height
  END Height;

PROCEDURE MaxEntries(VAR B : BTree):REAL;  (* Kapazitt eines Baumes *)
  BEGIN
    RETURN 2.0*FLOAT(MAX(CARDINAL))*FLOAT(B.n)
  END MaxEntries;

END BTREES.