MODULE ModList;

  FROM InOut IMPORT Read, Write, WriteLn, WriteCard, WriteString,
                    OpenInput, CloseInput, OpenOutput, CloseOutput, Done;

  CONST EOL = 36C; (* Wird meist von InOut oder ASCII exportiert *)
        CR  = 15C; (* Bewirkt auch Zeilenvorschub *)
        LF  = 12C; (* Wird ignoriert *)
        ESC = 33C;
        MaxZeichenZahl = 80;

  VAR Puffer : ARRAY[0..MaxZeichenZahl] OF CHAR;
      Zeichen : CHAR;
      Zustand : (ZeichenLesen, WortLesen, String1, String2, Kommentar);
      ZeilenNr, SeitenNr, ZeichenZahl, KommentarTiefe : CARDINAL;

(****************************************************************************)
(*        Lokales Modul zur gepufferten Eingabe eines Zeichens              *)
(****************************************************************************)

  MODULE GepuffertesLesen;

    IMPORT Read;
    EXPORT ReadChar, PushBack;

    VAR ZeichenPuffer : CHAR;

    PROCEDURE ReadChar(VAR Zeichen : CHAR);
      BEGIN
        IF ZeichenPuffer=0C
        THEN Read(Zeichen)
        ELSE Zeichen:=ZeichenPuffer; ZeichenPuffer:=0C
        END (* IF *)
      END ReadChar;

    PROCEDURE PushBack(Zeichen : CHAR);
      BEGIN
        ZeichenPuffer:=Zeichen
      END PushBack;

    BEGIN (* Initialisierung *)
      ZeichenPuffer:=0C
    END GepuffertesLesen;

(****************************************************************************)
(*        Lokales Modul zur Bestimmung, ob es sich bei einem Bezeichner     *)
(*        um ein Schluesselwort handelt.                                    *)
(****************************************************************************)

  MODULE ReservierteWoerter;

    EXPORT ReserviertesWort;

    VAR ResWort: ARRAY[1..40] OF ARRAY[0..15] OF CHAR;

    PROCEDURE ReserviertesWort(Bezeichner : ARRAY OF CHAR) : BOOLEAN;
      VAR von, bis, mitte : [1..40];

      PROCEDURE kleiner(X,Y : ARRAY OF CHAR) : BOOLEAN;
        VAR i, Minimum : CARDINAL;
        BEGIN
          IF HIGH(X)<HIGH(Y)
          THEN Minimum:=HIGH(X)
          ELSE Minimum:=HIGH(Y)
          END; (* IF *)
          i:=0;
          WHILE (i<Minimum) AND (X[i]=Y[i]) AND (X[i]<>0C) AND (Y[i]<>0C) DO
            INC(i)
          END;
          RETURN X[i]<Y[i]
        END kleiner;

      BEGIN
        von:=1; bis:=40;
        WHILE von<>bis DO
          mitte:=(von+bis) DIV 2;
          IF kleiner(ResWort[mitte],Bezeichner)
          THEN von:=mitte+1
          ELSE bis:=mitte
          END (* IF *)
        END; (* WHILE *)
        RETURN NOT kleiner(ResWort[von],Bezeichner) AND
               NOT kleiner(Bezeichner,ResWort[von])
      END ReserviertesWort;

    BEGIN
      ResWort[ 1]:="AND";               ResWort[21]:="LOOP";
      ResWort[ 2]:="ARRAY";             ResWort[22]:="MOD";
      ResWort[ 3]:="BEGIN";             ResWort[23]:="MODULE";
      ResWort[ 4]:="BY";                ResWort[24]:="NOT";
      ResWort[ 5]:="CASE";              ResWort[25]:="OF";
      ResWort[ 6]:="CONST";             ResWort[26]:="OR";
      ResWort[ 7]:="DEFINITION";        ResWort[27]:="POINTER";
      ResWort[ 8]:="DIV";               ResWort[28]:="PROCEDURE";
      ResWort[ 9]:="DO";                ResWort[29]:="QUALIFIED";
      ResWort[10]:="ELSE";              ResWort[30]:="RECORD";
      ResWort[11]:="ELSIF";             ResWort[31]:="REPEAT";
      ResWort[12]:="END";               ResWort[32]:="RETURN";
      ResWort[13]:="EXIT";              ResWort[33]:="SET";
      ResWort[14]:="EXPORT";            ResWort[34]:="THEN";
      ResWort[15]:="FOR";               ResWort[35]:="TO";
      ResWort[16]:="FROM";              ResWort[36]:="TYPE";
      ResWort[17]:="IF";                ResWort[37]:="UNTIL";
      ResWort[18]:="IMPLEMENTATION";    ResWort[38]:="VAR";
      ResWort[19]:="IMPORT";            ResWort[39]:="WHILE";
      ResWort[20]:="IN";                ResWort[40]:="WITH"
    END ReservierteWoerter;

(****************************************************************************)
(*        Die folgenden Prozeduren mssen fr den jeweils verwendeten       *)
(*        Drucker angepasst werden  (Hier Version fr Epson)                *)
(****************************************************************************)
  CONST ZeilenProSeite = 60;

  PROCEDURE InitDrucker;
    BEGIN
      Write(ESC); Write("M");
    END InitDrucker;

  PROCEDURE ExitDrucker;
    BEGIN
      Write(ESC); Write("P");
    END ExitDrucker;

  PROCEDURE MarkiereWort(Bezeichner : ARRAY OF CHAR);
    BEGIN
      Write(ESC); Write("E");
      WriteString(Bezeichner);
      Write(ESC); Write("F");
    END MarkiereWort;

  PROCEDURE KursivAn;
    BEGIN
      Write(ESC); Write("4");
    END KursivAn;

  PROCEDURE KursivAus;
    BEGIN
      Write(ESC); Write("5");
    END KursivAus;

(****************************************************************************)
(*                  Seitenumbruch und Zeilennummern                         *)
(****************************************************************************)

  PROCEDURE NeueSeite;
    BEGIN
      IF ZeilenNr>0 THEN Write(CHR(12)) END;
      INC(SeitenNr);
      KursivAn; WriteString("Seite "); WriteCard(SeitenNr,1); KursivAus;
      WriteLn; WriteLn
    END NeueSeite;

  PROCEDURE NeueZeile;
    BEGIN
      ReadChar(Zeichen); IF Zeichen<>LF THEN PushBack(Zeichen) END;
      IF ZeilenNr MOD ZeilenProSeite = 0
      THEN NeueSeite
      ELSE WriteLn
      END;
      INC(ZeilenNr); KursivAn; WriteCard(ZeilenNr,4); Write(':');
      IF KommentarTiefe=0 THEN KursivAus END
    END NeueZeile;

(****************************************************************************)
(*        Das Hauptprogramm als endlicher Automat                           *)
(****************************************************************************)

  BEGIN (* MODLIST *)
    WriteString("Quelltext-Lister"); WriteLn;
    WriteString("----------------"); WriteLn;
    WriteLn;
    OpenInput("MOD"); IF NOT Done THEN WriteString("Keine Datei!"); HALT END;
    OpenOutput("LST"); InitDrucker; KommentarTiefe:=0;
    ZeilenNr:=0; SeitenNr:=0; Zustand:=ZeichenLesen; Zeichen:=EOL;
    WHILE Done DO
      CASE Zustand OF
        ZeichenLesen : CASE Zeichen OF
                          'A'..'Z' : ZeichenZahl:=0;
                                     Puffer[ZeichenZahl]:=Zeichen;
                                     Zustand:=WortLesen
                        | '"'      : Write(Zeichen); Zustand:=String1
                        | "'"      : Write(Zeichen); Zustand:=String2
                        | "("      : ReadChar(Zeichen);
                                     IF Zeichen='*'
                                     THEN
                                       KursivAn;
                                       WriteString("(*");
                                       KommentarTiefe:=1;
                                       Zustand:=Kommentar
                                     ELSE
                                       Write("(");
                                       PushBack(Zeichen)
                                     END;
                        | EOL, CR  : NeueZeile;
                       ELSE Write(Zeichen);
                       END (* CASE *)
      | WortLesen    : IF (Zeichen>='A') AND (Zeichen<='Z')
                       THEN IF ZeichenZahl<MaxZeichenZahl
                            THEN
                              INC(ZeichenZahl);
                              Puffer[ZeichenZahl]:=Zeichen
                            END
                       ELSE (* Wort vollstaendig gelesen *)
                         IF ZeichenZahl<MaxZeichenZahl
                         THEN
                           INC(ZeichenZahl);
                           Puffer[ZeichenZahl]:=0C
                         END;
                         IF ReserviertesWort(Puffer)
                         THEN MarkiereWort(Puffer)
                         ELSE WriteString(Puffer)
                         END;
                         PushBack(Zeichen);
                         Zustand:=ZeichenLesen
                       END
      | String1      : Write(Zeichen);
                       IF Zeichen='"' THEN Zustand:=ZeichenLesen END
      | String2      : Write(Zeichen);
                       IF Zeichen="'" THEN Zustand:=ZeichenLesen END
      | Kommentar    : CASE Zeichen OF
                          '(' : ReadChar(Zeichen);
                                IF Zeichen='*'
                                THEN
                                  WriteString("(*");
                                  INC(KommentarTiefe)
                                ELSE
                                  Write("(");
                                  PushBack(Zeichen)
                                END;
                        | '*' : ReadChar(Zeichen);
                                IF Zeichen=')'
                                THEN
                                  WriteString("*)");
                                  DEC(KommentarTiefe);
                                  IF KommentarTiefe=0
                                  THEN
                                    KursivAus;
                                    Zustand:=ZeichenLesen
                                  END
                                ELSE
                                  Write("*");
                                  PushBack(Zeichen)
                                END;
                        | EOL, CR : NeueZeile
                       ELSE Write(Zeichen)
                       END (* CASE *)
      END; (* CASE *)
      ReadChar(Zeichen);
    END; (* WHILE *)
    CloseInput; ExitDrucker; CloseOutput;
  END ModList.