MODULE Othello;

FROM ChipIO1 IMPORT Write, WriteString, WriteCard, WriteLn, ReadKey, GotoXY,
                   LowVideo, HighVideo, ClrScr, ClrEol, WriteInt, ReadCard,
                   OpenWindow, OpenWindowHeader, CloseWindow, ScreenPtr,
                   SelWindow;

FROM BackTrack IMPORT MinMax;

FROM MENUE IMPORT Links, Rechts, ESC, ENTER, MenueString, Auswahl;

CONST KreuzZeichen = 'X';   (* Mensch *)
      KringelZeichen = 'O'; (* Computer *)
      LeerZeichen = ' ';
      MaxFelder = 8;

TYPE  Zustand = (Leer, Kreuz, Kringel);
      Feld = ARRAY[1..MaxFelder],[1..MaxFelder] OF Zustand;
      Zug = RECORD
              Spalte, Zeile : CARDINAL;
              Wert : INTEGER
            END;
      ZugListe  = RECORD
                    Anzahl : CARDINAL;
                    Zuege  : ARRAY[1..(MaxFelder*MaxFelder DIV 2)-2] OF Zug;
                  END;

VAR   Groesse, ZugNr, LeerFelder, Gesamt, Level : CARDINAL;
      Gegner :  ARRAY[Kreuz..Kringel] OF [Kreuz..Kringel];
      SpielFeld : Feld;
      SpielFenster, KreuzFenster, KringelFenster, AnzeigeFenster,
      ZugFenster, ParameterFenster : CARDINAL;
      Anzeige, SpielEnde, Abbruch : BOOLEAN;
      SpielStand : ARRAY[1..60] OF Feld;
      ZL : ZugListe;
      AmZug : [Kreuz..Kringel];

(******************************************************************************)
(**                  Die Bildschirmausgabe des Spielfelds                    **)
(******************************************************************************)
PROCEDURE SchreibSpielfeld;
 CONST  (* koennen an die jeweils gegebenen Moeglichkeitem angepasst werden *)
   LO = 332C; (* Ecke links oben    *)
   RO = 277C; (* Ecke rechts oben   *)
   SE = 263C; (* senkrechter Strich *)
   WA = 304C; (* waagrechter Strich *)
   LU = 300C; (* Ecke links unten   *)
   RU = 331C; (* Ecke rechts unten  *)
   SR = 303C; (* Lot nach rechts    *)
   SL = 264C; (* Lot nach links     *)
   SO = 301C; (* Lot nach oben      *)
   SU = 302C; (* Lot nach unten     *)
   KR = 305C; (* Kreuz              *)

 VAR i, j, x, y : CARDINAL;

 PROCEDURE LinieOben;
   VAR i : CARDINAL;
   BEGIN
     WriteString('  '); Write(LO);
     FOR i:=1 TO Groesse-1 DO
       Write(WA); Write(WA); Write(WA); Write(SU);
     END; (* FOR *)
     Write(WA); Write(WA); Write(WA); Write(RO); WriteLn;
   END LinieOben;

 PROCEDURE LinieUnten;
   VAR i : CARDINAL;
   BEGIN
     WriteString('  '); Write(LU);
     FOR i:=1 TO Groesse-1 DO
       Write(WA); Write(WA); Write(WA); Write(SO);
     END; (* FOR *)
     Write(WA); Write(WA); Write(WA); Write(RU); WriteString('  '); WriteLn
   END LinieUnten;

 BEGIN
   SelWindow(SpielFenster); LowVideo; x:=18-Groesse*2; y:=9-Groesse;
   GotoXY(x,y); Write(' ');
   FOR j:=1 TO Groesse DO WriteString('   '); Write(CHR(64+j)) END;
   INC(y); GotoXY(x,y); LinieOben;
   FOR i:=Groesse TO 1 BY -1 DO
     INC(y); GotoXY(x,y); WriteCard(i,2); Write(SE);
     FOR j:=1 TO Groesse DO
       WriteString(' '); HighVideo;
       CASE SpielFeld[j,i] OF
         Leer    : Write(LeerZeichen);
       | Kreuz   : Write(KreuzZeichen);
       | Kringel : Write(KringelZeichen);
       END; (* CASE *)
       LowVideo; Write(' '); Write(SE);
     END; (* FOR j *)
     WriteCard(i,1); INC(y); GotoXY(x,y);
     IF i>1
     THEN
       WriteString('  '); Write(SR);
       FOR j:=1 TO Groesse-1 DO Write(WA); Write(WA); Write(WA); Write(KR) END;
       Write(WA); Write(WA); Write(WA); Write(SL); WriteString('  '); WriteLn
     ELSE LinieUnten
     END; (* IF *)
   END; (* FOR *)
   INC(y); GotoXY(x,y); Write(' ');
   FOR j:=1 TO Groesse DO WriteString('   '); Write(CHR(64+j)) END;
   LowVideo
 END SchreibSpielfeld;

(****************************************************************************)
(*                   Umdrehen der Steine nach einem Zug                     *)
(****************************************************************************)
PROCEDURE FeldNeu(AmZug : Zustand; neuZug : Zug);
  TYPE  Richtung = [-1..1];
  VAR   i, j : Richtung;

  PROCEDURE Check(rx, ry : Richtung; neuZug : Zug);
     VAR S, Z : [0..MaxFelder];
         inRange : BOOLEAN;
         tPos : Zug;
     BEGIN
       tPos:=neuZug;
       WITH tPos DO
           REPEAT
             Spalte:=CARDINAL(INTEGER(Spalte)+rx);
             Zeile :=CARDINAL(INTEGER(Zeile) +ry);
             inRange:=(Spalte>=1) AND (Spalte<=Groesse) AND
                      (Zeile>=1) AND (Zeile<=Groesse);
           UNTIL NOT inRange OR (SpielFeld[Spalte,Zeile]<>Gegner[AmZug]);
           IF inRange AND (SpielFeld[Spalte,Zeile]=AmZug) THEN
             S:=CARDINAL(INTEGER(neuZug.Spalte)+rx);
             Z:=CARDINAL(INTEGER(neuZug.Zeile)+ry);
             WHILE (Spalte<>S) OR (Zeile<>Z) DO
               SpielFeld[S,Z]:=AmZug;
               S:=CARDINAL(INTEGER(S)+rx); Z:=CARDINAL(INTEGER(Z)+ry);
             END (* WHILE *)
           END (* IF *)
       END (* WITH *)
     END Check;

  BEGIN
    FOR i:=-1 TO 1 DO FOR j:=-1 TO 1 DO Check(i,j,neuZug) END END;
    SpielFeld[neuZug.Spalte,neuZug.Zeile]:=AmZug
  END FeldNeu;

(***************************************************************************)
(*                    Bewertung einer Stellung                             *)
(***************************************************************************)
TYPE KritischePositionen =
  ARRAY[1..4] OF RECORD
                   S,Z : CARDINAL;
                   Nachbarn : ARRAY[1..3] OF RECORD NS,NZ,Ft : CARDINAL END
                 END;
VAR KritPos : KritischePositionen;

PROCEDURE KeinZugMehrMoeglich():BOOLEAN; FORWARD;
(* Nur bei 1-Pass-Compilern notwendig *)

PROCEDURE InitKritPos;
  BEGIN
    WITH KritPos[1] DO
      S:=1; Z:=1;
      WITH Nachbarn[1] DO NS:=2; NZ:=2; Ft:=2 END;
      WITH Nachbarn[2] DO NS:=2; NZ:=1; Ft:=1 END;
      WITH Nachbarn[3] DO NS:=1; NZ:=2; Ft:=1 END;
    END;
    WITH KritPos[2] DO
      S:=1; Z:=Groesse;
      WITH Nachbarn[1] DO NS:=2; NZ:=Groesse-1; Ft:=2 END;
      WITH Nachbarn[2] DO NS:=2; NZ:=Groesse; Ft:=1 END;
      WITH Nachbarn[3] DO NS:=1; NZ:=Groesse-1; Ft:=1 END;
    END;
    WITH KritPos[3] DO
      S:=Groesse; Z:=1;
      WITH Nachbarn[1] DO NS:=Groesse-1; NZ:=2; Ft:=2; END;
      WITH Nachbarn[2] DO NS:=Groesse-1; NZ:=1; Ft:=1 END;
      WITH Nachbarn[3] DO NS:=Groesse; NZ:=2; Ft:=1 END;
    END;
    WITH KritPos[4] DO
      S:=Groesse; Z:=Groesse;
      WITH Nachbarn[1] DO NS:=Groesse-1; NZ:=Groesse-1; Ft:=2 END;
      WITH Nachbarn[2] DO NS:=Groesse-1; NZ:=Groesse; Ft:=1 END;
      WITH Nachbarn[3] DO NS:=Groesse; NZ:=Groesse-1; Ft:=1 END;
    END;
  END InitKritPos;

PROCEDURE StellungsWert() : INTEGER;
  VAR i, j : [1..MaxFelder]; b : Zustand; k, l : CARDINAL;
      Steine : ARRAY Zustand OF CARDINAL;
  BEGIN (* Wert *)
    FOR b:=MIN(Zustand) TO MAX(Zustand) DO Steine[b]:=0 END;
    FOR i:=1 TO Groesse DO
      FOR j:=1 TO Groesse DO
        INC(Steine[SpielFeld[i,j]]);
      END;
    END;
    IF (Steine[Leer]=0) OR (Steine[Kringel]=0) OR (Steine[Kreuz]=0)
       OR KeinZugMehrMoeglich()
    THEN
      IF Steine[Kreuz]>Steine[Kringel]
      THEN RETURN -1000+INTEGER(Steine[Kringel])-INTEGER(Steine[Kreuz])
      ELSIF Steine[Kringel]>Steine[Kreuz]
      THEN RETURN +1000+INTEGER(Steine[Kringel])-INTEGER(Steine[Kreuz])
      ELSE RETURN 0
      END
    ELSE
      FOR k:=1 TO 4 DO
        WITH KritPos[k] DO
          IF SpielFeld[S,Z]<>Leer
          THEN INC(Steine[SpielFeld[S,Z]],4*Steine[Leer])
          ELSE
            FOR l:=1 TO 3 DO
              WITH Nachbarn[l] DO
                IF SpielFeld[NS,NZ]<>Leer
                THEN INC(Steine[Gegner[SpielFeld[NS,NZ]]],Ft*Steine[Leer])
                END
              END (* WITH *)
            END (* FOR *)
          END (* IF *)
        END (* WITH *)
      END (* FOR *)
    END; (* IF *)
    RETURN INTEGER(Steine[Kringel])-INTEGER(Steine[Kreuz]);
  END StellungsWert;

(****************************************************************************)
(*                      Der Zuglisten-Generator                             *)
(****************************************************************************)
PROCEDURE GenZugListe(AnDerReihe : Zustand; VAR ZL : ZugListe;
                      Test, SortOption : BOOLEAN);
 TYPE Richtung = [-1..1];
 VAR   erlaubt  : BOOLEAN;
       i, j     : Richtung;
       S, Z     : [1..MaxFelder];
       e        : BOOLEAN;
       SF       : Feld;
       k        : CARDINAL;

PROCEDURE Sort(VAR ZL : ZugListe); (* einfaches Sortieren durch Auswahl *)
  VAR i,j,m : CARDINAL;
      t : Zug;
  BEGIN (* Sort *)
    WITH ZL DO
      FOR i:=1 TO Anzahl-1 DO
        m:=i;
        FOR j:=i+1 TO Anzahl DO
          IF Zuege[j].Wert>Zuege[m].Wert THEN m:=j END;
        END; (* FOR *)
        IF m<>i THEN t:=Zuege[i]; Zuege[i]:=Zuege[m]; Zuege[m]:=t END
      END (* FOR *)
    END (* WITH *)
  END Sort;

 PROCEDURE check(rx, ry : Richtung; SX, SY : CARDINAL);
   VAR x,y,s,z : CARDINAL;
       geg : BOOLEAN;
   BEGIN
     x:=SX; y:=SY;
     REPEAT
       x:=CARDINAL(INTEGER(x)+rx); y:=CARDINAL(INTEGER(y)+ry);
       e:=(x>0) AND (x<=Groesse) AND (y>0) AND (y<=Groesse);
       IF e THEN geg:=(SpielFeld[x,y]<>Gegner[AnDerReihe]) END
     UNTIL NOT e OR geg;
     IF NOT geg OR NOT e THEN RETURN END;
     s:=x; z:=y;
     IF (s>0) AND (s<=Groesse) AND (z>0) AND (z<=Groesse) AND
        (SpielFeld[s,z]=AnDerReihe)
     THEN
        SX:=CARDINAL(INTEGER(SX)+rx); SY:=CARDINAL(INTEGER(SY)+ry);
        IF (SX<>s) OR (SY<>z) THEN erlaubt:=TRUE END
     END
   END check;

 BEGIN (* GenZugListe *)
   WITH ZL DO
      Anzahl:=0;
      FOR S:=1 TO Groesse DO
        FOR Z:=1 TO Groesse DO
          IF SpielFeld[S,Z]=Leer THEN
            erlaubt:=FALSE;
            FOR i:=-1 TO 1 DO
              FOR j:=-1 TO 1 DO
                IF (i<>0) OR (j<>0) THEN check(i,j,S,Z) END
              END (* FOR j *)
            END; (* FOR i *)
            IF erlaubt THEN
              INC(Anzahl); WITH Zuege[Anzahl] DO Spalte:=S; Zeile:=Z END;
              IF Test THEN RETURN END
            END (* IF erlaubt *)
          END (* IF SpielFeld[S,Z]=leer *)
        END (* FOR Z *)
      END; (* FOR S *)
      IF (Anzahl>1) AND SortOption THEN
        SF:=SpielFeld;
        FOR k:=1 TO Anzahl DO
          WITH Zuege[k] DO
            SpielFeld:=SF; FeldNeu(AnDerReihe,Zuege[k]);
            IF AnDerReihe=Kringel
            THEN Wert:=StellungsWert()
            ELSE Wert:=-StellungsWert()
            END
          END (* WITH *)
        END; (* FOR *)
        SpielFeld:=SF;
        Sort(ZL)
      END; (* IF *)
    END; (* WITH *)
  END GenZugListe;

PROCEDURE KeinZugMehrMoeglich():BOOLEAN;
  VAR ZL : ZugListe;
  BEGIN
    GenZugListe(Kringel,ZL,TRUE,FALSE);
    IF ZL.Anzahl<>0 THEN RETURN FALSE END;
    GenZugListe(Kreuz,ZL,TRUE,FALSE);
    RETURN ZL.Anzahl=0
  END KeinZugMehrMoeglich;

PROCEDURE AnzahlSteine(Wer : Zustand):CARDINAL;
  VAR i,j,z : CARDINAL;
  BEGIN
    z:=0;
    FOR i:=1 TO Groesse DO
      FOR j:=1 TO Groesse DO
        IF SpielFeld[i,j]=Wer THEN INC(z) END
      END
    END;
    RETURN z
  END AnzahlSteine;

PROCEDURE SchreibZugListe(AmZug : Zustand; ZL : ZugListe);
  VAR i, t : CARDINAL;
  BEGIN
    t:=ScreenPtr; SelWindow(ZugFenster); GotoXY(1,1); ClrEol;
    IF ZL.Anzahl>0 THEN
      IF AmZug=Kreuz THEN Write(KreuzZeichen) ELSE Write(KringelZeichen) END;
      Write(" ");
      WITH ZL DO
        FOR i:=1 TO Anzahl DO
          Write(CHR(64+Zuege[i].Spalte));
          Write(CHR(48+Zuege[i].Zeile));
          IF i<Anzahl THEN Write("/") END
        END (* FOR *)
      END (* WITH *)
    END; (* IF *)
    SelWindow(t);
  END SchreibZugListe;

PROCEDURE SchreibSteine;
  VAR t : CARDINAL;
  BEGIN
    t:=ScreenPtr; SelWindow(KringelFenster); GotoXY(11,1);
    HighVideo; WriteCard(AnzahlSteine(Kringel),2);
    SelWindow(KreuzFenster); GotoXY(11,1);
    HighVideo; WriteCard(AnzahlSteine(Kreuz),2); SelWindow(t)
  END SchreibSteine;

(****************************************************************************)
(*                     Eingabe der Spielparameter                           *)
(****************************************************************************)
PROCEDURE InitSpiel;
  VAR i, j, m, t : CARDINAL;
  BEGIN
    InitKritPos; ZugNr:=0; SpielEnde:=FALSE; t:=ScreenPtr;
    FOR i:=1 TO Groesse DO FOR j:=1 TO Groesse DO SpielFeld[i,j]:=Leer END END;
    Gegner[Kreuz]:=Kringel; Gegner[Kringel]:=Kreuz; m:=Groesse DIV 2;
    SpielFeld[m,m]:=Kreuz; SpielFeld[m+1,m+1]:=Kreuz;
    SpielFeld[m,m+1]:=Kringel; SpielFeld[m+1,m]:=Kringel;
    SelWindow(SpielFenster); ClrScr; SchreibSpielfeld;
    SelWindow(KringelFenster); GotoXY(12,1); ClrEol; GotoXY(12,2); ClrEol;
    SelWindow(KreuzFenster); GotoXY(12,1); ClrEol; GotoXY(12,2); ClrEol;
    SelWindow(AnzeigeFenster); ClrScr; SelWindow(ZugFenster); ClrScr;
    SelWindow(t)
  END InitSpiel;

PROCEDURE ParameterEingabe;

  MODULE Init;
    IMPORT MenueString;
    EXPORT Parameter, FeldGroesse, Anfang;
    VAR Parameter : ARRAY[1..7] OF MenueString;
        FeldGroesse : ARRAY[1..3] OF MenueString;
        Anfang : ARRAY[1..2] OF MenueString;
    BEGIN
      Parameter[1]:="Zum Spiel"; Parameter[2]:="Neues Spiel";
      Parameter[3]:="Level";     Parameter[4]:="Rcknahme";
      Parameter[5]:="Anzeige";   Parameter[6]:="Spielfeld";
      Parameter[7]:="Ende";
      FeldGroesse[1]:="klein";   FeldGroesse[2]:="mittel";
      FeldGroesse[3]:="gro";
      Anfang[1]:="Mensch";       Anfang[2]:="Computer";
    END Init;

  VAR Wahl, t, G1, Anfaenger : CARDINAL;
      Zeichen, C : CHAR;

  PROCEDURE SchreibParameter;
    VAR t : CARDINAL;
    BEGIN
      t:=ScreenPtr; SelWindow(ParameterFenster); HighVideo;
      GotoXY(16,3); Write(CHR(Level+48));
      GotoXY(16,5);
      IF Anzeige THEN WriteString("an ") ELSE WriteString("aus") END;
      GotoXY(16,6);
      CASE Groesse OF
        4 : WriteString("klein ");
      | 6 : WriteString("mittel");
      | 8 : WriteString("gro  ")
      END;
      LowVideo; SelWindow(t);
    END SchreibParameter;

  BEGIN
    t:=ScreenPtr; Wahl:=1; SchreibParameter; SelWindow(SpielFenster);
    REPEAT
      SelWindow(ParameterFenster); GotoXY(20,1); WriteString("Men aktiv");
      Auswahl(2,1,1,1,7,Parameter,"ZNLRASE",TRUE,TRUE,Wahl,Zeichen);
      IF (Zeichen<>Links) AND (Zeichen<>Rechts) AND (Zeichen<>ESC) THEN
        CASE Wahl OF
          1 : (* Nichts *)
        | 2 : IF (ZugNr=0) OR SpielEnde
              THEN InitSpiel
              ELSE
                OpenWindow(61,10,80,17);
                WriteString("Soll das laufende"); WriteLn;
                WriteString("Spiel wirklich"); WriteLn;
                WriteString("abgebrochen"); WriteLn;
                WriteString("werden (J/N)? ");
                IF CAP(ReadKey())="J"
                THEN CloseWindow; InitSpiel; Abbruch:=TRUE;
                ELSE CloseWindow
                END
              END;
        | 3 : GotoXY(16,3); HighVideo;
              REPEAT C:=ReadKey() UNTIL (C=ESC) OR (C>="1") AND (C<="9");
              IF C<>ESC THEN Level:=ORD(C)-ORD("0") END;
              Write(CHR(Level+48)); LowVideo;
        | 4 : IF ZugNr>1
              THEN
                DEC(ZugNr); SpielEnde:=FALSE; AmZug:=Kreuz;
                SpielFeld:=SpielStand[ZugNr]; SchreibSpielfeld;
                GenZugListe(Kreuz,ZL,FALSE,TRUE);
                SchreibZugListe(Kreuz,ZL); SchreibSteine
              END;
        | 5 : IF Anzeige THEN SelWindow(AnzeigeFenster); ClrScr END;
              Anzeige:=NOT Anzeige; SchreibParameter;
        | 6 : IF ZugNr=0
              THEN
                OpenWindow(60,6,80,10); G1:=(Groesse-2)DIV 2;
                Auswahl(2,1,1,1,3,FeldGroesse,"kmg",TRUE,TRUE,G1,C);
                CloseWindow;
                IF C<>ESC THEN Groesse:=G1*2+2; SchreibParameter; InitSpiel END;
              ELSE
                OpenWindow(61,10,80,17);
                WriteString("Sie mssen erst"); WriteLn;
                WriteString("das Spiel beenden"); WriteLn;
                WriteString("oder den Punkt"); WriteLn;
                WriteString("<Neues Spiel>"); WriteLn;
                WriteString("whlen. (ESC)"); WriteLn;
                WHILE ReadKey()<>ESC DO END;
                CloseWindow;
              END;
        | 7 : OpenWindow(61,10,80,17);
              WriteString("Wollen Sie"); WriteLn;
              WriteString("wirklich das"); WriteLn;
              WriteString("Spiel beenden"); WriteLn;
              WriteString("(J/N)? ");
              IF CAP(ReadKey()) = "J"
              THEN SelWindow(0); ClrScr; HALT
              ELSE CloseWindow
              END
        END
      END
    UNTIL (Wahl=1) AND ((Zeichen=ENTER) OR (CAP(Zeichen)="Z"));
    GotoXY(20,1); ClrEol;
    IF ZugNr=0 THEN
      OpenWindow(61,10,80,17);
      WriteString("Wer soll"); WriteLn;
      WriteString("beginnen?"); Anfaenger:=1;
      Auswahl(3,4,1,1,2,Anfang,"MC",TRUE,FALSE,Anfaenger,C);
      CloseWindow;
      AmZug:=VAL(Zustand,Anfaenger)
    END;
    SelWindow(t)
  END ParameterEingabe;

(****************************************************************************)
(*                         Zug des Menschen                                 *)
(****************************************************************************)
MODULE MenschenEingabe;

  IMPORT Zustand, SpielFeld, Gegner, Zug, ZugListe, Abbruch,
         FeldNeu, GotoXY, ClrEol, Write, WriteString, ReadKey,
         GenZugListe, KreuzZeichen, SelWindow, KreuzFenster, Groesse,
         ParameterEingabe, ZL, ESC, HighVideo;
  EXPORT MenschenZug;

   PROCEDURE LegalerZug(Z : Zug):BOOLEAN;
     VAR i : CARDINAL;
     BEGIN
       FOR i:=1 TO ZL.Anzahl DO
         IF (ZL.Zuege[i].Spalte=Z.Spalte) AND (ZL.Zuege[i].Zeile=Z.Zeile)
         THEN RETURN TRUE
         END
       END; (* FOR *)
       RETURN FALSE
     END LegalerZug;

  PROCEDURE MenschenZug;
    VAR C : CHAR;
        tZug : Zug;
    BEGIN
      GenZugListe(Kreuz,ZL,FALSE,TRUE); SelWindow(KreuzFenster);
      IF ZL.Anzahl=0
      THEN WriteString(" nicht mglich!")
      ELSE
        REPEAT
          REPEAT
            GotoXY(12,2); ClrEol; C:=ReadKey(); C:=CAP(C); HighVideo;
            IF C=ESC
            THEN ParameterEingabe; IF Abbruch THEN RETURN END;
            ELSE Write(C) END;
          UNTIL (C > CHR(64)) AND (C<=CHR(64+Groesse));
          tZug.Spalte:=ORD(C)-64;
          REPEAT
            GotoXY(13,2); ClrEol; C:=ReadKey(); C:=CAP(C); HighVideo;
            IF C=ESC
            THEN ParameterEingabe; IF Abbruch THEN RETURN END;
            ELSE Write(C) END;
          UNTIL (C > CHR(48)) AND (C<=CHR(48+Groesse));
          tZug.Zeile:=ORD(C)-48;
        UNTIL LegalerZug(tZug);
        FeldNeu(Kreuz,tZug);
      END
    END MenschenZug;

  END MenschenEingabe;

(****************************************************************************)
(*                     Zug des Computers                                    *)
(****************************************************************************)
MODULE ComputerLogik;

  IMPORT MinMax, Zustand, SpielFeld, Gegner, StellungsWert, ClrScr,
         Zug, ZugListe, GenZugListe, Feld, FeldNeu, GotoXY, Write, WriteInt,
         ClrEol, Level, Anzeige, WriteString, SelWindow, KringelFenster,
         ZugFenster, AnzeigeFenster, HighVideo, LowVideo, ScreenPtr;

  EXPORT ZugTiefe, ComputerZug;

  CONST MaxTiefe = 12;

  TYPE SpielStand = RECORD
                      ErsterZug  : BOOLEAN;  (* Informationen fuer *)
                      LetzterZug : Zug; (* den Zuggenerator *)
                      Spiel      : Feld;
                      ZugNr      : CARDINAL;
                      ZL         : ZugListe;
                      AmZug      : [Kreuz..Kringel]
                    END;

  VAR StkPtr       : CARDINAL;
      SpielStack   : ARRAY [1..MaxTiefe] OF SpielStand;
      BestMoves    : ARRAY [1..MaxTiefe], [1..MaxTiefe] OF Zug;
      Markiert     : ARRAY [1..MaxTiefe] OF CARDINAL;
      ZugTiefe     : CARDINAL;

  PROCEDURE InitSpielStack;
    BEGIN
      StkPtr:=1;
      SpielStack[1].AmZug:=Kringel;
      Markiert[1]:=0;
    END InitSpielStack;

  PROCEDURE ZugZurueck;
    BEGIN
      DEC(StkPtr); SpielFeld:=SpielStack[StkPtr].Spiel;
    END ZugZurueck;

  PROCEDURE MarkiereZug(W : INTEGER);
    VAR i, t : CARDINAL;
    BEGIN
      BestMoves[StkPtr,1]:=SpielStack[StkPtr].LetzterZug;
      FOR i:=1 TO Markiert[StkPtr+1] DO
        BestMoves[StkPtr,i+1]:=BestMoves[StkPtr+1,i]
      END;
      Markiert[StkPtr]:=1+Markiert[StkPtr+1];
      IF Anzeige AND (StkPtr=1) THEN
        t:=ScreenPtr; SelWindow(AnzeigeFenster); LowVideo; GotoXY(1,1); ClrEol;
        FOR i:=1 TO Markiert[1] DO
          WITH BestMoves[1,i] DO
            Write(CHR(64+Spalte)); Write(CHR(48+Zeile)); Write("/")
          END
        END;
        GotoXY(1,2); WriteString("Bewertung: "); WriteInt(W,1); ClrEol;
        SelWindow(t);
      END
    END MarkiereZug;

  PROCEDURE ComputerAmZug() : BOOLEAN;
    BEGIN
      WITH SpielStack[StkPtr] DO
        GenZugListe(AmZug,ZL,FALSE,StkPtr<ZugTiefe);
        IF ZL.Anzahl=0 THEN
          AmZug:=Gegner[AmZug];
          GenZugListe(AmZug,ZL,FALSE,StkPtr<ZugTiefe);
        END; (* IF *)
        ErsterZug:=TRUE;
        RETURN AmZug=Kringel
      END; (* WITH *)
    END ComputerAmZug;

  PROCEDURE NaechsterZug() : BOOLEAN;
    VAR t : CARDINAL;
    BEGIN
      WITH SpielStack[StkPtr] DO
        IF ErsterZug THEN ZugNr:=1; ErsterZug:=FALSE ELSE INC(ZugNr) END;
        IF ZugNr>ZL.Anzahl
        THEN RETURN FALSE
        ELSE
          Spiel:=SpielFeld; LetzterZug:=ZL.Zuege[ZugNr];
          FeldNeu(AmZug,LetzterZug);
          SpielStack[StkPtr+1].AmZug:=Gegner[AmZug];
          IF StkPtr=1
          THEN
            t:=ScreenPtr; SelWindow(ZugFenster); HighVideo;
            GotoXY(3+(ZugNr-1)*3,1);
            WITH LetzterZug DO Write(CHR(Spalte+64)); Write(CHR(Zeile+48)) END;
            LowVideo; SelWindow(t);
          END
        END; (* IF *)
      END; (* WITH *)
      INC(StkPtr); Markiert[StkPtr]:=0; RETURN TRUE;
    END NaechsterZug;

  PROCEDURE ComputerZug;
    VAR W : INTEGER;
    BEGIN
      SelWindow(AnzeigeFenster); ClrScr;
      SelWindow(KringelFenster); GotoXY(12,2); ClrEol;
      InitSpielStack;
      W:=MinMax(ZugTiefe,ComputerAmZug,NaechsterZug,ZugZurueck,
                MarkiereZug, StellungsWert);
      IF Markiert[1]=0
      THEN Write("!"); HALT
      ELSE
        HighVideo;
        WITH BestMoves[1,1] DO Write(CHR(Spalte+64)); Write(CHR(Zeile+48)) END;
        LowVideo; FeldNeu(Kringel,BestMoves[1,1]);
      END
    END ComputerZug;

END ComputerLogik;

(****************************************************************************)
(*                        Das Hauptprogramm                                 *)
(****************************************************************************)

BEGIN
  Groesse:=8; ClrScr; Anzeige:=FALSE; Level:=4;
  OpenWindowHeader(1,1,41,21,"[ Spielfeld ]"); SpielFenster:=ScreenPtr;
  AmZug:=Kreuz; SpielEnde:=FALSE;
  OpenWindowHeader(45,10,60,13,"[ Mensch ]"); KreuzFenster:=ScreenPtr;
  WriteString(" Steine:"); WriteLn; WriteString(" Ihr Zug: ");
  OpenWindowHeader(45,14,60,17,"[ Computer ]"); KringelFenster:=ScreenPtr;
  WriteString(" Steine:"); WriteLn; WriteString(" Mein Zug: ");
  OpenWindowHeader(1,22,80,24,"[ Mgliche Zge ]"); ZugFenster:=ScreenPtr;
  OpenWindowHeader(45,18,80,21,"[ Voraussichtliche Entwicklung ]");
  AnzeigeFenster:=ScreenPtr;
  OpenWindowHeader(45,1,80,9,"[ OTHELLO 2.0 ]");
  ParameterFenster:=ScreenPtr; InitSpiel; Abbruch:=FALSE;
  LOOP
    ParameterEingabe;
    REPEAT
      LeerFelder:=AnzahlSteine(Leer); Gesamt:=Groesse*Groesse;
      IF LeerFelder<10 THEN ZugTiefe:=10
      ELSIF LeerFelder<Gesamt DIV 3 THEN ZugTiefe:=Level+2
      ELSIF (Level>2) AND (LeerFelder>2*Gesamt DIV 3) THEN ZugTiefe:=Level-2
      ELSE ZugTiefe:=Level
      END;
      SchreibSteine;
      GenZugListe(AmZug,ZL,FALSE,TRUE); SchreibZugListe(AmZug,ZL);
      IF ZL.Anzahl=0
      THEN
        AmZug:=Gegner[AmZug];
        GenZugListe(AmZug,ZL,FALSE,TRUE);
        IF ZL.Anzahl=0
        THEN SpielEnde:=TRUE
        ELSE
          OpenWindow(61,10,80,17);
          IF Gegner[AmZug]=Kreuz
          THEN WriteString("Sie knnen")
          ELSE WriteString("Ich kann")
          END;
          WriteString(" nicht!"); WriteLn; WriteLn;
          WriteString("Bitte ESC drcken");
          WHILE ReadKey()<>ESC DO END;
          CloseWindow;
          SchreibZugListe(AmZug,ZL);
        END
      END;
      IF NOT SpielEnde
      THEN
        IF AmZug=Kreuz
        THEN INC(ZugNr); SpielStand[ZugNr]:=SpielFeld; MenschenZug
        ELSE ComputerZug END;
        SchreibSpielfeld;
        IF Abbruch THEN Abbruch:=FALSE ELSE AmZug:=Gegner[AmZug] END
      ELSE
        OpenWindow(61,10,80,17);
        IF AnzahlSteine(Kreuz)>AnzahlSteine(Kringel)
        THEN
          WriteString("Gratuliere!"); WriteLn;
          WriteString("Sie haben"); WriteLn;
          WriteString("gewonnen.");
        ELSIF AnzahlSteine(Kreuz)<AnzahlSteine(Kringel)
        THEN
          WriteString("Das war's dann."); WriteLn;
          WriteString("Der Sieg ist mein!");
        ELSE
          WriteString("Glck gehabt!"); WriteLn;
          WriteString("Unentschieden");
        END;
        WriteLn; WriteLn;
        WriteString("Weiter mit ESC ");
        WHILE ReadKey()<>ESC DO END;
        CloseWindow;
      END;
    UNTIL SpielEnde;
  END (* LOOP *)
END Othello.