MODULE TTT;

  FROM ChipIO IMPORT
      WriteString, WriteLn, WriteInt, WriteCard,
      Write, ReadString, ReadLn,
      GotoXY, ClrEol, ClrScr;

  FROM BackTrack IMPORT MinMax;

  CONST MaxTiefe = 10;

  TYPE Zustand    = (Leer, Computer, Gegner);
       Feld       = ARRAY [1..3], [1..3] OF Zustand;
       SpielZug   = RECORD
                      X, Y : CARDINAL; (* Kurznotation fuer den Zug *)
                    END;
       SpielStand = RECORD
                      ErsterZug  : BOOLEAN;  (* Informationen fuer *)
                      LetzterZug : SpielZug; (* den Zuggenerator *)

                      (* Optional, fuer einfache Zugruecknahme *)
                      Spiel      : Feld;

                      (* Weitere Informationen fuer *)
                      (* den Zuggenerator koennen folgen *)
                    END;

  VAR Level     : CARDINAL;
      SpielFeld : Feld;
      Stack     : ARRAY [1..MaxTiefe] OF SpielStand;
      Zuege     : ARRAY [1..MaxTiefe], [1..MaxTiefe] OF SpielZug;
      Markiert  : ARRAY [1..MaxTiefe] OF CARDINAL;

  PROCEDURE GrundStellung;
    VAR X,Y : CARDINAL;
    BEGIN
      FOR X:=1 TO 3 DO          (* Grundstellung aufbauen *)
        FOR Y:=1 TO 3 DO
          SpielFeld[X,Y]:=Leer
        END
      END;
    END GrundStellung;

  PROCEDURE InitStack;
    VAR X,Y : CARDINAL;
    BEGIN
      Markiert[1]:=0;           (* MarkiereZug initialisieren       *)
      Level:=1;                 (* Stack-Level initialisieren       *)
    END InitStack;

  PROCEDURE ZugZurueck;
    BEGIN
      DEC(Level);
      SpielFeld:=Stack[Level].Spiel;
    END ZugZurueck;

  PROCEDURE MarkiereZug(M : INTEGER); (* M wird hier nicht verwendet *)
    VAR i : CARDINAL;
    BEGIN
      Zuege[Level,1]:=Stack[Level].LetzterZug;
      FOR i:=1 TO Markiert[Level+1] DO
        Zuege[Level,i+1]:=Zuege[Level+1,i]
      END;
      Markiert[Level]:=1+Markiert[Level+1]
    END MarkiereZug;

  (* Falls man nicht an einer Zug-Voraussage interessiert ist:
  PROCEDURE MarkiereZug(M : INTEGER);
    BEGIN
      IF Level=1 THEN
        BesterZug:=Stack[1].LetzterZug
      END
    END MarkiereZug;
  *)

  PROCEDURE ComputerAmZug() : BOOLEAN;
    BEGIN
      Stack[Level].ErsterZug:=TRUE;
      RETURN ODD(Level);
    END ComputerAmZug;

  PROCEDURE StellungsWert() : INTEGER;
    VAR X,Y : CARDINAL;
    BEGIN
        IF (SpielFeld[1,1]=SpielFeld[2,2]) AND
           (SpielFeld[2,2]=SpielFeld[3,3]) AND
           (SpielFeld[3,3]<>Leer)
        THEN
          IF SpielFeld[3,3]=Computer
          THEN RETURN INTEGER(1+MaxTiefe-Level)
          ELSE RETURN -INTEGER(1+MaxTiefe-Level)
          END
        ELSIF (SpielFeld[1,3]=SpielFeld[2,2]) AND
              (SpielFeld[2,2]=SpielFeld[3,1]) AND
              (SpielFeld[3,1]<>Leer)
        THEN
          IF SpielFeld[3,1]=Computer
          THEN RETURN INTEGER(1+MaxTiefe-Level)
          ELSE RETURN -INTEGER(1+MaxTiefe-Level)
          END
        ELSE
          FOR X:=1 TO 3 DO
            IF (SpielFeld[X,1]=SpielFeld[X,2]) AND
               (SpielFeld[X,2]=SpielFeld[X,3]) AND
               (SpielFeld[X,1]<>Leer)
            THEN
              IF SpielFeld[X,1]=Computer
              THEN RETURN INTEGER(1+MaxTiefe-Level)
              ELSE RETURN -INTEGER(1+MaxTiefe-Level)
              END
            END
          END;
          FOR Y:=1 TO 3 DO
            IF (SpielFeld[1,Y]=SpielFeld[2,Y]) AND
               (SpielFeld[2,Y]=SpielFeld[3,Y]) AND
               (SpielFeld[1,Y]<>Leer)
            THEN
              IF SpielFeld[1,Y]=Computer
              THEN RETURN INTEGER(1+MaxTiefe-Level)
              ELSE RETURN -INTEGER(1+MaxTiefe-Level)
              END
            END
          END
        END;
      RETURN 0;
      (* Wenn nur +-1 als Ergebnis genommen werden,
         spielt der Computer sehr eigenartig! *)
    END StellungsWert;

  PROCEDURE NaechsterZug() : BOOLEAN;
    BEGIN
      WITH Stack[Level] DO WITH LetzterZug DO
        REPEAT
          IF ErsterZug
          THEN
            IF StellungsWert()<>0
            THEN
              RETURN FALSE;
            END;
            X:=1; Y:=1; ErsterZug:=FALSE
          ELSE
            IF (X=3) AND (Y=3)
            THEN
              RETURN FALSE;
            ELSIF (X=3)
            THEN
              X:=1; INC(Y);
            ELSE
              INC(X);
            END;
          END;
        UNTIL SpielFeld[X,Y]=Leer;
        Spiel:=SpielFeld;
        IF ODD(Level)
        THEN SpielFeld[X,Y]:=Computer
        ELSE SpielFeld[X,Y]:=Gegner
        END
      END END;
      INC(Level);
      Markiert[Level]:=0;
      RETURN TRUE;
    END NaechsterZug;

PROCEDURE SchreibSpielFeld;
  VAR X, Y : CARDINAL;
  BEGIN
    GotoXY(1,1);
    WriteString("   A  B  C "); WriteLn;
    FOR Y:=3 TO 1 BY -1 DO
      WriteCard(Y,2);
      FOR X:=1 TO 3 DO
        CASE SpielFeld[X,Y] OF
          Leer     : WriteString("   ")|
          Gegner   : WriteString(" X ")|
          Computer : WriteString(" O ")
        END; (* CASE *)
      END; (* FOR X *)
      WriteLn;
    END; (* FOR Y *)
    WriteLn;
  END SchreibSpielFeld;

PROCEDURE SpielEnde() : BOOLEAN;
  VAR X, Y : CARDINAL;
  BEGIN
    IF
      StellungsWert()=0
    THEN
      FOR Y:=1 TO 3 DO
        FOR X:=1 TO 3 DO
          IF SpielFeld[X,Y]=Leer THEN
            RETURN FALSE
          END;
        END; (* FOR X *)
      END; (* FOR Y *)
    END; (* IF *)
    RETURN TRUE
  END SpielEnde;

VAR Eingabe : ARRAY [0..1] OF CHAR;
    X, Y, Z : CARDINAL;
    W : INTEGER;
    Anfang : BOOLEAN;
BEGIN
  WriteString("Tic Tac Toe"); WriteLn;
  WriteString("==========="); WriteLn; WriteLn;
  WriteString("Wollen Sie anfangen ? ");
  ReadString(Eingabe);
  Anfang:=CAP(Eingabe[0])<>'N';
  ClrScr;
  GrundStellung;
  SchreibSpielFeld;
  REPEAT
    IF Anfang THEN
      REPEAT
        REPEAT
          GotoXY(1,7);
          WriteString("Ihr Zug ? "); ClrEol;
          ReadString(Eingabe); ReadLn;
          Eingabe[0]:=CAP(Eingabe[0]);
        UNTIL (Eingabe[0]>="A") AND (Eingabe[0]<="C")
              AND (Eingabe[1]>="1") AND (Eingabe[1]<="3");
        X:=ORD(Eingabe[0])-ORD("A")+1;
        Y:=ORD(Eingabe[1])-ORD("1")+1;
      UNTIL SpielFeld[X,Y]=Leer;
      SpielFeld[X,Y]:=Gegner;
      SchreibSpielFeld;
    END;
    Anfang:=TRUE;
    GotoXY(1,10); ClrEol; WriteLn; ClrEol;
    GotoXY(1,9);
    WriteString("Mein Zug : "); ClrEol;
    InitStack;
    W:=MinMax(9,ComputerAmZug,NaechsterZug,
              ZugZurueck,MarkiereZug,StellungsWert);
    IF Markiert[1]<>0
    THEN
      Write(CHR(Zuege[1,1].X-1+ORD("A")));
      Write(CHR(Zuege[1,1].Y-1+ORD("1")));
      SpielFeld[Zuege[1,1].X,Zuege[1,1].Y]:=Computer;
    END;
    WriteLn;
    WriteString("Wert : "); WriteInt(W,1); WriteLn;
    WriteString("Zugvorausschau : ");
    FOR Z:=2 TO Markiert[1] DO
      Write(CHR(Zuege[1,Z].X-1+ORD("A")));
      Write(CHR(Zuege[1,Z].Y-1+ORD("1")));
      Write("/");
    END;
    SchreibSpielFeld;
  UNTIL SpielEnde();
  GotoXY(1,24);
END TTT.